"""Command."""


# [ Imports ]
import logging


# [ Logging ]
logger = logging.getLogger(__name__)


# [ API ]
# XXX Better doc strings
# XXX add unit tests
class Command:
    """Command."""

    def __init__(
        self,
        command: str,
        *args: str
    ) -> None:
        """Init the state."""
        self.command = command
        self.args = args



"""Source Control."""


# [ Imports ]
# [ -Python ]
from pathlib import Path
import logging
import sys
from typing import Sequence
import concurrent.futures
import asyncio
import select
import termios
# # [ -Third Party ]
# import blessed
# [ -Project ]
from ..interactors.runner import run


# [ Logging ]
logger = logging.getLogger(__name__)


# XXX Better doc strings
# XXX add unit tests
# [ Helpers ]
async def _get_status() -> Sequence[str]:
    """Get repo status."""
    logger.info('getting status...')
    return (await run('git', ['status', '--porcelain'])).output.splitlines()


async def _add_missing_paths(paths: Sequence[Path], status_lines: Sequence[str]) -> None:
    """Add any missing paths to the repo."""
    logger.info('adding missing paths...')
    if not status_lines:
        return
    new_file_lines = [l for l in status_lines if l.startswith('??')]
    new_path_strings = [v for l in new_file_lines for k, v in [l.split()]]
    for path in paths:
        path_string = str(path)
        if path_string in new_path_strings or (
            any(nps.endswith('/') and path_string.startswith(nps) for nps in new_path_strings)
        ):
            await run('git', ['add', path_string])


async def async_input(prompt: str) -> str:
    """Async input prompt."""
    readable = []  # type: List[int]
    print(prompt, end='')
    sys.stdout.flush()
    while not readable:
        readable, writeable, executable = select.select([sys.stdin], [], [], 0)
        try:
            await asyncio.sleep(0.1)
        except concurrent.futures.CancelledError:
            print("input cancelled...")
            termios.tcflush(sys.stdin, termios.TCIFLUSH)
            raise
    return sys.stdin.readline().rstrip()


# [ API ]
async def commit(paths: Sequence[Path]) -> bool:
    """Commit the current repo."""
    logger.info('committing...')
    status_lines = await _get_status()
    await _add_missing_paths(paths, status_lines)
    status_lines = await _get_status()
    modified_file_lines = [l for l in status_lines if not l.startswith('??')]
    if not modified_file_lines:
        print("No actual changes made - nothing to commit.")
        return True
    print((await run('git', ['diff', '--color'])).output)
    try:
        print("Files are still being watched, and further changes will cancel this dialog")
        print("  and re-run the checks, but you may commit the above changes now by")
        print("  entering a commit message.")
        print()
        message = await async_input('> ')
        # XXX use selectors to wait for input to be ready to grab.
    except concurrent.futures.CancelledError:
        raise
    return (await run('git', ['commit', '-am', message])).success



"""Factory for real file paths."""


# [ Imports ]
from pathlib import Path
import logging
from typing import Any


# [ Pyflakes ]
assert Any


# [ Logging ]
logger = logging.getLogger(__name__)


# XXX Better doc strings
# XXX add unit tests
# [ API ]
def real_file_path(path_string: str) -> Path:
    """Build an absolute, concrete, existing path to a file."""
    path = Path(path_string)  # type: Any
    if not path.is_absolute():
        path = Path.cwd() / path_string
    if not path.is_file():
        raise FileNotFoundError("Specified path is not a file: {}".format(str(path)))
    return path



"""Run commands."""


# [ Imports ]
# [ -Python ]
import subprocess
from types import SimpleNamespace
from typing import Sequence
import asyncio
import logging
import concurrent.futures


# [ Logging ]
logger = logging.getLogger(__name__)


# [ Helpers ]
# XXX better doc strings
# XXX add unit tests
async def run(command: str, args: Sequence[str]) -> SimpleNamespace:
    """Run the command."""
    logger.info("Running command: {} {}".format(command, ' '.join(args)))
    process = await asyncio.create_subprocess_exec(
        command, *args,
        stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
    )
    # XXX way too much nesting here
    try:
        await process.wait()
    except concurrent.futures.CancelledError:
        process.terminate()
        try:
            await asyncio.wait_for(process.wait(), timeout=3)
        # XXX mypy says no such error
        except asyncio.TimeoutError:  # type: ignore
            process.kill()
            try:
                await asyncio.wait_for(process.wait(), timeout=3)
            # XXX mypy says no such error
            except asyncio.TimeoutError:  # type: ignore
                raise RuntimeError("subprocess for {} did not stop after terminate and kill commands.".format(command))
        raise
    return SimpleNamespace(
        success=process.returncode == 0,
        output=str(await process.stdout.read(), 'utf-8')
    )



"""Check the files."""


# [ Imports ]
# [ -Python ]
import logging
from typing import Sequence, Callable, Awaitable, List, Iterable
from pathlib import Path
from types import SimpleNamespace
import asyncio
import concurrent.futures
# [ -Third Party ]
import a_sync
# [ -Project ]
from ..interactors.runner import run
from ..thin_types.command import Command


# [ Logging ]
logger = logging.getLogger(__name__)


# [ Helpers ]
def report_result(command: str, result: SimpleNamespace) -> None:
    """Report the result."""
    outcome = 'pass' if result.success else 'fail'
    print('result for {}: {}'.format(command, outcome))
    print("{} output:".format(command))
    print(result.output)


async def _run_single(
    command: Command,
    all_path_strings: Sequence[str],
    changed_path_strings: Sequence[str]
) -> SimpleNamespace:
    """Run single command."""
    # XXX abstract the substitution stuff out
    args = command.args
    all_symbol = '{all}'
    changed_symbol = '{changed}'
    substitution_occurred = False
    while all_symbol in args:
        all_index = args.index(all_symbol)
        before = args[:all_index]
        after = args[all_index + 1:]
        args = list(before) + list(all_path_strings) + list(after)
        substitution_occurred = True
    while changed_symbol in args:
        changed_index = args.index(changed_symbol)
        before = args[:changed_index]
        after = args[changed_index + 1:]
        args = list(before) + list(changed_path_strings) + list(after)
        substitution_occurred = True
    if not substitution_occurred:
        args += all_path_strings

    print("running {}...".format(command.command))
    try:
        result = await run(command.command, args)
    except concurrent.futures.CancelledError:
        print("result for {}: cancelled".format(command.command))
        raise
    except FileNotFoundError:
        result = SimpleNamespace(
            success=False,
            output="cannot run command ({}) - no such executable found.".format(command.command)
        )
    report_result(command.command, result)
    return result


async def _cancel_checks(pending: Iterable[asyncio.Future]) -> None:
    """Cancel the given checks."""
    if pending and not all(p.done() for p in pending):
        print("Cancelling running checks...")
        # XXX mypy says there is no gather
        gathered = asyncio.gather(*pending)  # type: ignore
        gathered.cancel()
        # XXX shield from cancellation, since we're waiting for cancellation completion?
        await asyncio.wait_for(gathered, timeout=None)


async def _run_parallel_checks(tasks: List[asyncio.Task]) -> bool:
    """Run the checks in parallel."""
    all_done = False
    while not all_done:
        done, pending = await asyncio.wait(tasks, return_when=concurrent.futures.FIRST_COMPLETED)
        if not all(f.result().success for f in done):
            await _cancel_checks(pending)
            return False
        all_done = not pending
    return True


# [ API ]
# XXX Better doc strings
# XXX colored check/x
# XXX bold current command
# XXX normal old command
# XXX add unit tests
class Checker:
    """Checker."""

    def __init__(
        self, *,
        get_commands: Callable[[], Awaitable[Sequence[Command]]],
        get_paths: Callable[[], Awaitable[Sequence[Path]]],
        on_success: Callable[[Sequence[Path]], Awaitable[bool]]
    ) -> None:
        """Init the state."""
        self._get_commands = get_commands
        self._get_paths = get_paths
        self._unsafe_on_success = on_success
        self._changed_path_strings = []  # type: List[str]

    async def _on_success(self, paths: Sequence[Path]) -> bool:
        """Run the passed in 'on_success' function safely."""
        try:
            result = await a_sync.run(self._unsafe_on_success, paths)
        except concurrent.futures.CancelledError:
            logger.info("success task cancelled - new changes detected.")
            raise
        except Exception:
            logger.exception("Unexpected exception while running 'on_success' callback from checker.")
            exit(1)
        return result

    async def run(self, changed_paths) -> None:
        """Run the checks."""
        logger.info("running the static checkers...")

        paths = await self._get_paths()
        path_strings = tuple(str(p) for p in paths)
        for the_path in changed_paths:
            the_path_string = str(the_path)
            if (
                (the_path_string not in self._changed_path_strings) and
                the_path_string in path_strings
            ):
                self._changed_path_strings.append(the_path_string)

        commands = await self._get_commands()
        tasks = []
        try:
            for command in commands:
                # XXX mypy says asyncio doesn't have ensure_future
                task = asyncio.ensure_future(_run_single(command, path_strings, self._changed_path_strings))  # type: ignore
                tasks.append(task)
            all_passed = await _run_parallel_checks(tasks)
        except concurrent.futures.CancelledError:
            logger.info("checking cancelled - cancelling running checks")
            await _cancel_checks(tasks)
            raise

        success_action_succeeded = False
        if all_passed:
            success_action_succeeded = await self._on_success(paths)
        if success_action_succeeded:
            self._changed_path_strings = []



"""CLI args."""


# [ Imports ]
# [ -Python ]
import logging
from typing import Any
from argparse import ArgumentParser
# [ -Third Party ]
import a_sync
# [ -Project ]
from ..thin_types.real_file_path import real_file_path


# [ Logging ]
logger = logging.getLogger(__name__)


# XXX Better doc strings
# XXX match globs and directories
# [ Helpers ]
def get_parser() -> ArgumentParser:
    """Get arg parser."""
    parser = ArgumentParser()
    parser.add_argument(
        '-c', '--config',
        help="Path to the yaml config file to use. [default: %(default)s]",
        type=real_file_path,
        default='check.yaml'
    )
    parser.add_argument(
        '-1', '--once',
        help="Only run the checks once [default: False]",
        action='store_true'
    )
    return parser


# [ API ]
async def get_arg(arg: str) -> Any:
    """Return the arg."""
    parser = get_parser()
    try:
        args = await a_sync.run(parser.parse_args)
    except Exception as e:
        # XXX mypy says this is incompatible - it is not.
        parser.error(e)  # type: ignore
    return getattr(args, arg)



"""Setup for pocketwalk."""


# [ Imports ]
# [ -Python ]
from setuptools import setup, find_packages


# [ Main ]
setup(
    name='pocketwalk',
    version='0.1.0',
    description='Pocketwalk: Watch/Run/Commit.',
    url='https://github.com/toejough/pocketwalk',
    author='toejough',
    author_email='toejough@gmail.com',
    classifiers=[
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            'Development Status :: 3 - Alpha',

            # Indicate who your project is intended for
            'Intended Audience :: Developers',

            # Pick your license as you wish (should match "license" above)
            'License :: OSI Approved :: Apache Software License',

            # Specify the Python versions you support here. In particular, ensure
            # that you indicate whether you support Python 2, Python 3 or both.
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5',
    ],
    keywords="watch run commit",
    packages=find_packages(),
    install_requires=['a-sync', 'blessed', 'pyaml'],
    entry_points={
        'console_scripts': [
            'pocketwalk=pocketwalk:main',
        ],
    }
)



"""Static checking."""


# [ Imports ]
# [ -Python ]
from pathlib import Path
import logging
from typing import Sequence
import asyncio
import signal
import concurrent.futures
# XXX mypy drops some error about where glob is defined?
import glob  # type: ignore
import itertools
# [ -Third Party ]
import a_sync
# [ -Project ]
from .components.checker import Checker
from .components.watcher import Watcher
from .components.source_control import commit
from .interactors import config
from .interactors import cli
from .thin_types.command import Command


# [ Logging ]
logging.basicConfig(level=logging.WARN)
logger = logging.getLogger(__name__)


# XXX Better doc strings
# XXX make logging level configurable
# XXX make logging location configurable
# XXX add unit tests
# [ Helpers ]
async def _get_watch_paths() -> Sequence[Path]:
    """Return the paths to watch."""
    check_paths = await _get_check_paths()
    config_path = await cli.get_arg('config')
    return [*check_paths, config_path]


async def _get_check_paths() -> Sequence[Path]:
    """Return the paths to check."""
    path_strings = await config.get_config('paths')
    # XXX mypy says no 'recursive' arg for glob.
    unglobbed_path_strings = []
    for this_path_string in path_strings:
        unglobbed_path_list = glob.glob(this_path_string, recursive=True)  # type: ignore
        unglobbed_path_strings.append(unglobbed_path_list)
    flattened_path_strings = itertools.chain(*unglobbed_path_strings)
    paths = [Path(p) for p in flattened_path_strings]
    return paths


async def _get_commands() -> Sequence[Command]:
    """Return the commands to run."""
    return [Command(*c) for c in await config.get_config('checks')]


async def _check(checker: Checker) -> None:
    """Check the files."""
    if await cli.get_arg('once'):
        await checker.run(_get_watch_paths())
    else:
        await Watcher(
            get_paths=_get_watch_paths,
            on_modification=checker.run
        ).watch()


def handler() -> None:
    """Handle SIGINT by cancelling everything."""
    print('\nCaught Ctrl-c.')
    all_tasks = list(asyncio.Task.all_tasks())
    pending_tasks = [t for t in all_tasks if not t.done()]
    if pending_tasks:
        print('Winding down...')
        for task in pending_tasks:
            task.cancel()


async def _unfriendly_main() -> None:
    """Not ctrl-c-friendly."""
    logger.info('Running static checker...')

    checker = Checker(
        get_commands=_get_commands,
        get_paths=_get_check_paths,
        on_success=commit
    )

    await _check(checker)


# [ Main ]
def main() -> None:
    """Interruptable main func."""
    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal.SIGINT, handler)
    try:
        a_sync.block(_unfriendly_main)
    except concurrent.futures.CancelledError:
        pass
    finally:
        all_tasks = list(asyncio.Task.all_tasks())
        pending_tasks = [t for t in all_tasks if not t.done()]
        loop = asyncio.get_event_loop()
        if pending_tasks:
            loop.run_until_complete(asyncio.wait(pending_tasks))
        loop.close()
        print("Done.")


# satisfy vulture:
assert main



"""Watch files."""


# [ Imports ]
# [ -Python ]
import logging
from pprint import pformat
from typing import Sequence, Dict, Callable, Awaitable, Optional, Any
from pathlib import Path
import asyncio
import concurrent.futures
# [ -Third Party ]
import blessed
import a_sync


# [ Terminal ]
T = blessed.Terminal()


# [ Logging ]
logger = logging.getLogger(__name__)


# [ Helpers ]
# XXX Better doc strings
# XXX add unit tests
async def _get_mtimes(paths: Sequence[Path]) -> Dict[Path, float]:
    """Return the mtimes for the paths."""
    logger.debug("getting mtimes for {}".format(pformat(paths)))

    mtimes = {}
    for p in paths:
        try:
            stats = await a_sync.run(p.stat)
            mtimes[p] = stats.st_mtime
        except FileNotFoundError:
            # Try again?
            # fail out if not found a second time.
            await asyncio.sleep(0.1)
            stats = await a_sync.run(p.stat)
            mtimes[p] = stats.st_mtime
    return mtimes


async def _wait(*, silently: bool) -> None:
    """Wait."""
    if not silently:
        print("\rWaiting for changes...{}".format(
            T.clear_eol
        ), end='')
    await asyncio.sleep(1)


def _get_changed_paths(last_mtimes: Dict[Path, float], new_mtimes: Dict[Path, float]) -> Sequence[Path]:
    """Get the changed paths."""
    # XXX actually only get the changed paths - verify content changes before returning
    # XXX ^ will require a separate, special run the very first time (run all, regardless of changes)
    new_paths = [k for k in new_mtimes if k not in last_mtimes]
    maintained_paths = [k for k in last_mtimes if k in new_mtimes]
    modified_paths = [p for p in maintained_paths if new_mtimes[p] != last_mtimes[p]]
    removed_paths = [k for k in last_mtimes if k not in new_mtimes]
    return new_paths + modified_paths + removed_paths


# [ API ]
class Watcher:
    """Watch files."""

    def __init__(
        self, *,
        get_paths: Callable[[], Awaitable[Sequence[Path]]],
        # XXX clarify and remove unnecessary Any types
        on_modification: Callable[[Any], Optional[Awaitable[None]]]
    ) -> None:
        """Init the state."""
        self._get_paths = get_paths
        self._unsafe_on_modification = on_modification

    async def _on_modification(self, changed_paths: Sequence[Path]) -> None:
        """Safe version of _unsafe_on_modification."""
        try:
            await a_sync.run(self._unsafe_on_modification, changed_paths)
        except concurrent.futures.CancelledError:
            logger.info("Run-checks cancelled.")
            raise
        except Exception:
            logger.exception("Unexpected exception while running 'on_modification' callback from watcher.")
            exit(1)

    async def watch(self) -> None:
        """Watch the paths."""
        last_mtimes = {}  # type: Dict[Path, float]
        running_checks = None
        # XXX way too complex
        try:
            while True:
                paths = await self._get_paths()
                new_mtimes = await _get_mtimes(paths)
                changed_paths = _get_changed_paths(last_mtimes, new_mtimes)
                if changed_paths:
                    print("\rFound changes in files:{}\n{}".format(
                        T.clear_eol,
                        pformat([str(p) for p in changed_paths])
                    ))
                    if running_checks:
                        if running_checks.cancel():
                            print("Cancelled running checks due to mid-check changes.")
                    # XXX mypy says ensure_future is not a thing
                    running_checks = asyncio.ensure_future(self._on_modification(changed_paths))  # type: ignore
                    last_mtimes = new_mtimes
                else:
                    silently = running_checks and not running_checks.done()
                    await _wait(silently=silently)
                    # await a_sync.run(_wait, silently=silently)
        except concurrent.futures.CancelledError:
            print("Stopped watching files.")
            raise



"""Config help."""


# [ Imports ]
# [ -Python ]
import logging
from typing import Any
# [ -Third Party ]
import a_sync
import yaml
# [ -Project ]
from . import cli


# [ Logging ]
logger = logging.getLogger(__name__)


# XXX Better doc strings
# XXX rename get_config to just get
# XXX add unit tests
# [ Helper Helpers ]
async def get_config(section: str) -> Any:
    """Return the config."""
    config_path = await cli.get_arg('config')
    # XXX pull out just the unsafe stuff
    try:
        return (await a_sync.run(yaml.load, config_path.read_text()))[section]
    except FileNotFoundError:
        logger.critical(
            "The config file ({}) was not found.  Cannot continue.".format(
                str(config_path)
            ),
            stack_info=True
        )
        exit(1)
